/*Copyright (c) 2015-2050, JD All rights reserved.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License. */

/* --------------------------------------------------
 * @brief: 源文件 一个设备示例
 *
 * @version: 1.0
 *
 * --------------------------------------------------
 */

#include <stdio.h>
#include "joylink_device.h"
#include "joylink_string.h"
#include "joylink_memory.h"
#include "joylink_log.h"
#include "joylink_time.h"
#include "joylink_thread.h"

#include "joylink_protocol.h"
#include "joylink_sdk.h"

#include "cJSON.h"
#include "joylink_device_ota.h"
#include "joylink_sub_device.h"
#include "joylink_agent.h"

// 示例属性实例名
#define PROP_KEY        ("light.switch")

// 示例事件实例名、事件参数实例名
#define EVENT_KEY       ("light.tips")
#define EVENT_ARG_KEY   ("time")

/**
 * @brief  示例设备结构体定义 和 静态变量定义
 * 
 */
typedef struct {
    struct joylink_sdk_t *sdk;
    int32_t is_connected;
    uint16_t  prop_version;
    bool switch_button; // properties
} device_demo_t;
static device_demo_t device_demo;

// 静态函数定义
static int32_t device_thing_model_post(struct joylink_sdk_t *sdk);
static int32_t device_properties_post(struct joylink_sdk_t *sdk);
static int32_t device_event_tips_post(struct joylink_sdk_t *sdk);
static int32_t device_ota_properties_post(struct joylink_sdk_t *sdk);

/**
 * @brief  设备初始化
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 */
int32_t device_initialise(struct joylink_sdk_t *sdk)
{
    jl_platform_memset(&device_demo, 0, sizeof(device_demo_t));
    device_demo.sdk = sdk;
    device_demo.is_connected = 0;
    device_demo.prop_version = 0;
    device_demo.switch_button = false;

    // 初始化子设备
    sub_device_initialise(sdk);

    // 设备任务线程
    jl_thread_t task_id;
    task_id.thread_task = (threadtask) device_task_thread;
	task_id.stackSize = 0x4000;
    task_id.priority = JL_THREAD_PRI_DEFAULT;
    task_id.parameter = NULL;
    jl_platform_thread_start(&task_id);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  连接成功 回调函数
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 *    sdk连接京东云成功后，包括初次连接以及其它原因导致的断网重连，都会调用该函数
 */
int32_t device_callback_connect(struct joylink_sdk_t *sdk)
{
    // 物模型上报
    device_thing_model_post(sdk);

    // 上报自身属性
    device_properties_post(sdk);

    // 设备OTA属性上报
    device_ota_properties_post(sdk);

    // 设备OTA版本号上报
    char ota_version[32];
    device_ota_get_version(ota_version);
    joylink_sdk_dev_ota_version_changed_post(sdk, "device", NULL, ota_version);

    // 连接代理可以选择在物管平台建立设备的拓扑绑定关系
    // 也可以在网管侧主动进行设备拓扑绑定关系的维护
    // 连接代理上报子设备绑定状态
    // agent_event_topology_add_devices(sdk);
 
    // 连接代理上报子设备解绑状态
    // agent_event_topology_delete_devices(sdk);

    // 代理子设备进行上线状态上报
    joylink_sdk_sub_dev_online_post(sdk, "device", sub_device_get_info(NULL));

    // 用户可根据自身情况进行逻辑处理
    device_demo.is_connected = 1;

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  断开连接 回调函数
 *
 * @param[in] sdk: joylink sdk 结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 *    sdk断开连接，包括主动断开，由其它原因导致的被动断开，都会调用该函数
 */
int32_t device_callback_disconnect(struct joylink_sdk_t *sdk)
{
    // 用户可根据自身情况进行逻辑处理
    device_demo.is_connected = 0;

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  设备物模型信息上报
 * 
 * @note
 * 
 */
static int32_t device_thing_model_post(struct joylink_sdk_t *sdk)
{
    // 物模型上报 示例代码
    JoyProtoThingModelPost_t post;
    jl_platform_memset(&post, 0, sizeof(JoyProtoThingModelPost_t));
    post.deviceId = SDK_CONFIG_DEVICE_ID;
    post.messageId = joylink_sdk_get_messageId(NULL);
    post.timestamp = jl_platform_time(NULL);
    post.thing_model.id = DEVICE_INFO_THING_MODEL_ID;
    post.thing_model.version = DEVICE_INFO_THING_MODEL_VERSION;
    joylink_sdk_dev_thing_model_post(sdk, JOYLINK_PROTO_OBJ_NAME_DEV, NULL, &post);
    jl_platform_free(post.messageId);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  设备属性上报
 * 
 * @note
 * 
 */
static int32_t device_properties_post(struct joylink_sdk_t *sdk)
{
    // 连接成功 属性上报
    JoyProtoPropPost_t prop_post;
    jl_platform_memset(&prop_post, 0, sizeof(JoyProtoPropPost_t));
    prop_post.deviceId = SDK_CONFIG_DEVICE_ID;
    prop_post.timestamp = jl_platform_time(NULL);
    prop_post.version = device_demo.prop_version;
    prop_post.messageId = joylink_sdk_get_messageId(NULL);

    prop_post.prop_num = 1;
    prop_post.properties = (JoyProtoKeyValue_t **) jl_platform_malloc(prop_post.prop_num * sizeof(JoyProtoKeyValue_t *));
    jl_platform_memset(prop_post.properties, 0, prop_post.prop_num * sizeof(JoyProtoKeyValue_t *));
    prop_post.properties[0] = (JoyProtoKeyValue_t *) jl_platform_malloc(prop_post.prop_num * sizeof(JoyProtoKeyValue_t));
    prop_post.properties[0]->key = (char *) jl_platform_malloc(32 * sizeof(char));
    jl_platform_strcpy(prop_post.properties[0]->key, PROP_KEY);
    prop_post.properties[0]->value = joylink_proto_keyvalue_pack_bool(device_demo.switch_button);

    joylink_sdk_dev_prop_post(sdk, JOYLINK_PROTO_OBJ_NAME_DEV, NULL, &prop_post);
    prop_post.deviceId = NULL;
    joylink_proto_free_prop_post(&prop_post);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  设备事件上报(示例事件)
 * 
 * @note
 * 
 */
static int32_t device_event_tips_post(struct joylink_sdk_t *sdk)
{
    jl_time_stamp_t time_stamp;
    char date[32];

    JoyProtoEvtPost_t evt_post;
    jl_platform_memset(&evt_post, 0, sizeof(JoyProtoEvtPost_t));
    evt_post.deviceId = SDK_CONFIG_DEVICE_ID;
    evt_post.timestamp = jl_platform_time(NULL);
    evt_post.messageId = joylink_sdk_get_messageId(NULL);
    evt_post.evt_num = 1;
    evt_post.events = (JoyProtoEvtPostEvt_t **) jl_platform_malloc (evt_post.evt_num * sizeof(JoyProtoEvtPostEvt_t *));
    evt_post.events[0] = (JoyProtoEvtPostEvt_t *) jl_platform_malloc (sizeof(JoyProtoEvtPostEvt_t));
    jl_platform_memset(evt_post.events[0], 0, sizeof(JoyProtoEvtPostEvt_t));
    evt_post.events[0]->key = (char *) jl_platform_malloc(32 * sizeof(char));
    jl_platform_strcpy(evt_post.events[0]->key, EVENT_KEY);
    evt_post.events[0]->param_num = 1;
    evt_post.events[0]->parameters = (JoyProtoKeyValue_t **) jl_platform_malloc(evt_post.events[0]->param_num * sizeof(JoyProtoKeyValue_t *));
    evt_post.events[0]->parameters[0] = (JoyProtoKeyValue_t *) jl_platform_malloc(sizeof(JoyProtoKeyValue_t));
    jl_platform_memset(evt_post.events[0]->parameters[0], 0, sizeof(JoyProtoKeyValue_t));
    evt_post.events[0]->parameters[0]->key = (char *) jl_platform_malloc(32 * sizeof(char));
    jl_platform_strcpy(evt_post.events[0]->parameters[0]->key, EVENT_ARG_KEY);

    jl_platform_time_get_timestamp(&time_stamp);
    jl_platform_sprintf(date, "%d%d", time_stamp.second, time_stamp.ms);
    evt_post.events[0]->parameters[0]->value = joylink_proto_keyvalue_pack_string(date);

    // 上报事件
    joylink_sdk_dev_evt_post(sdk, JOYLINK_PROTO_OBJ_NAME_DEV, NULL, &evt_post);

    // 释放资源
    evt_post.deviceId = NULL;
    joylink_proto_free_evt_post(&evt_post);

}

/**
 * @brief  设备OTA属性上报
 * 
 * @note
 * 
 */
static int32_t device_ota_properties_post(struct joylink_sdk_t *sdk)
{
    // 连接成功 ota 属性上报
    JoyProtoPropPost_t prop_post;
    jl_platform_memset(&prop_post, 0, sizeof(JoyProtoPropPost_t));
    prop_post.deviceId = SDK_CONFIG_DEVICE_ID;
    prop_post.timestamp = jl_platform_time(NULL);
    prop_post.version = 0;
    prop_post.messageId = joylink_sdk_get_messageId(NULL);
    prop_post.properties = device_ota_get_properties(&prop_post.prop_num);

    joylink_sdk_dev_prop_post(sdk, JOYLINK_PROTO_OBJ_NAME_DEV, JOYLINK_PROTO_SERVICE_OTA, &prop_post);
    prop_post.deviceId = NULL;
    joylink_proto_free_prop_post(&prop_post);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  物模型上报响应 回调函数
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] obj_name: 对象空间，可选值{"device", "edge", "group"}
 * @param[in] service_key: 服务实例名字
 * @param[in] in_res: 物模型上报响应消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 *    sdk 进行无模型上报后，收到上报响应后 会调用该函数
 */
int32_t device_callback_thingmodel_post_response(struct joylink_sdk_t *sdk,
                        char *obj_name, char *service_key, JoyProtoThingModelPostRes_t *in_res)
{
    // 对象空间(obj_name) 仅支持 "device"，建议实现时进行一下判别
    // 服务实例名，对于物模型上报响应来说，目前为空，开发者可以暂时不用进行逻辑判断
    
    /* 示例代码
     * if (in_res->code != 200) {log_error("failed")};
     */

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  属性设置 回调函数
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] obj_name: 对象空间，可选值{"device", "edge", "group"}
 * @param[in] service_key: 服务实例名字
 * @param[in] in_set: 属性设置消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 *    sdk 收到属性设置消息时 会调用该函数
 */
int32_t device_callback_property_set(struct joylink_sdk_t *sdk,
                        char *obj_name, char *service_key, JoyProtoPropSet_t *in_set)
{
    /*
     * 属性值 in_set->properties[0]->value 为json字符串，sdk 提供了与bool，int32_t，double和string转换的接口,
     * 其它类型数据，请自行根据定义进行解析
     * 例如：从value中解析出int32_t可调用：
     *     joylink_proto_keyvalue_parse_int32(char *in_value, int32_t *out_val);
     * 如果想将int32_t 打包成 value 可调用：
     *     char *joylink_proto_keyvalue_pack_int32(int32_t in_int32);
     */ 

    // 发送 属性设置响应示例代码
    JoyProtoPropSetRes_t out_res;
    jl_platform_memset(&out_res, 0, sizeof(JoyProtoPropSetRes_t));
    out_res.deviceId = in_set->deviceId;
    out_res.messageId = in_set->messageId;
    out_res.timestamp = jl_platform_time(NULL);
    
    // 检查请求版本号
    if (joylink_sdk_prop_version_check(device_demo.prop_version, in_set->version)) {
        out_res.code = 200;
    }
    else {
        out_res.code = 204;
    }
    
    if (out_res.code == 200) {
        // 检查属性实例名
        if (!jl_platform_strcmp(in_set->properties[0]->key, PROP_KEY)) {
            // 解析期望的值
            joylink_proto_keyvalue_parse_bool(in_set->properties[0]->value, &device_demo.switch_button);
            
            // 更新版本号
            device_demo.prop_version = in_set->version;
            
            // 给出响应
            out_res.prop_num = 1;
            out_res.properties = (JoyProtoKeyValue_t **) jl_platform_malloc(out_res.prop_num * sizeof(JoyProtoKeyValue_t *));
            jl_platform_memset(out_res.properties, 0, out_res.prop_num * sizeof(JoyProtoKeyValue_t *));
            out_res.properties[0] = (JoyProtoKeyValue_t *) jl_platform_malloc(sizeof(JoyProtoKeyValue_t));
            out_res.properties[0]->key = (char *) jl_platform_malloc (32 * sizeof(char));
            jl_platform_strcpy(out_res.properties[0]->key, PROP_KEY);
            out_res.properties[0]->value = joylink_proto_keyvalue_pack_bool(device_demo.switch_button);
            out_res.version = device_demo.prop_version;
        }
        else {
            out_res.code = 400;
        }
    }

    joylink_sdk_dev_prop_set_response(sdk, obj_name, service_key, &out_res);
    
    out_res.deviceId = NULL;
    out_res.messageId = NULL;
    joylink_proto_free_prop_set_res(&out_res);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  属性获取 回调函数
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] obj_name: 对象空间，可选值{"device", "edge", "group"}
 * @param[in] service_key: 服务实例名字
 * @param[in] in_get: 属性获取消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 *    sdk 收到属性设置消息时 会调用该函数
 */
int32_t device_callback_property_get(struct joylink_sdk_t *sdk,
                        char *obj_name, char *service_key, JoyProtoPropGet_t *in_get)
{
    // 处理流程及属性值(properties[0]->value) 参见属性设置的说明

    // 发送 属性设置响应示例代码
    JoyProtoPropGetRes_t out_res;
    jl_platform_memset(&out_res, 0, sizeof(JoyProtoPropGetRes_t));
    out_res.deviceId = in_get->deviceId;
    out_res.messageId = in_get->messageId;
    out_res.timestamp = jl_platform_time(NULL);
    /*
     * out_res.properties[0]->value = joylink_proto_keyvalue_pack_int32(200);
     * 注意内存释放问题
     */
    out_res.prop_num = 1;
    out_res.properties = (JoyProtoKeyValue_t **) jl_platform_malloc(out_res.prop_num * sizeof(JoyProtoKeyValue_t *));
    jl_platform_memset(out_res.properties, 0, out_res.prop_num * sizeof(JoyProtoKeyValue_t *));
    out_res.properties[0] = (JoyProtoKeyValue_t *) jl_platform_malloc(sizeof(JoyProtoKeyValue_t));
    out_res.properties[0]->key = (char *) jl_platform_malloc (32 * sizeof(char));
    jl_platform_strcpy(out_res.properties[0]->key, PROP_KEY);
    out_res.properties[0]->value = joylink_proto_keyvalue_pack_bool(device_demo.switch_button);
    
    out_res.version = device_demo.prop_version;
    out_res.code = 200;

    joylink_sdk_dev_prop_get_response(sdk, obj_name, service_key, &out_res);
    
    out_res.deviceId = NULL;
    out_res.messageId = NULL;
    joylink_proto_free_prop_get_res(&out_res);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  方法调用 回调函数
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] obj_name: 对象空间，可选值{"device", "edge", "group"}
 * @param[in] service_key: 服务实例名字
 * @param[in] in_call: 方法调用消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 *    sdk 收到方法调用消息时 会调用该函数
 */
int32_t device_callback_function_call(struct joylink_sdk_t *sdk,
                        char *obj_name, char *service_key, JoyProtoFuncCall_t *in_call)
{
    // 方法调用入参及出参 参数值(in[0]->value) 与属性值相同为 json 字符串

    // 发送 属性设置响应示例代码
    JoyProtoFuncCallRes_t out_res;
    jl_platform_memset(&out_res, 0, sizeof(JoyProtoFuncCallRes_t));
    out_res.deviceId = in_call->deviceId;
    out_res.messageId = in_call->messageId;
    /*
     * out_res.functions[0]->out[0]->value = joylink_proto_keyvalue_pack_int32(200);
     * 注意内存释放问题
     */
    joylink_sdk_dev_func_call_response(sdk, obj_name, service_key, &out_res);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  NTP请求响应 回调函数
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] set_time: 设置时间的结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 *    sdk 收到NTP请求响应 会调用该函数
 */
int32_t device_callback_ntp_response(struct joylink_sdk_t *sdk, jl_time_stamp_t *set_time)
{
    // TODO 如果有需要，进行时钟同步

    log_error("NTP Time Needed to be Set: %d.%d", set_time->second, set_time->ms);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  设备任务线程函数
 * 
 * @note
 *    设备自己的任务线程
 */
void device_task_thread(void)
{
    // 任务循环, 只是一个示例，来说明API的调用方法，真是流程请根据具体的业务场景确定
    while (true)
    {
        if (device_demo.is_connected) {
            // 网关设备属性周期上报
            // device_properties_post(device_demo.sdk);
            // jl_platform_msleep(5000);

            // 上报自身事件
            // device_event_tips_post(device_demo.sdk);
            // jl_platform_msleep(5000);

            // 代理子设备上报属性
            // sub_device_properties_post(device_demo.sdk, NULL);
            // jl_platform_msleep(5000);

            // 代理子设备上报事件
            // sub_device_event_post(device_demo.sdk, NULL);
        }

        // time wait
        jl_platform_msleep(20000);
    }
}

// end of file
