/*Copyright (c) 2015-2050, JD All rights reserved.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License. */

/* --------------------------------------------------
 * @brief: 头文件 设备ota示例
 *
 * @version: 1.0
 *
 * --------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include "joylink_device.h"
#include "joylink_string.h"
#include "joylink_memory.h"
#include "joylink_log.h"
#include "joylink_time.h"
#include "joylink_thread.h"

#include "joylink_protocol.h"
#include "joylink_sdk.h"

#include "joylink_device_ota.h"
#include "cJSON.h"

// ota 相关的静态变量
static char ota_url[512] = {0};
static char ota_task_id[128] = {0};
static char ota_install_model[8] = {0};
static char ota_package_name[256] = {};
static char ota_version[32] = {"v1.0.0"};
static char ota_task_version[32] = {0};
static uint32_t ota_task_exec_time = 0;

// 当前状态 错误码 进度
static int32_t ota_state = JOY_OTA_ST_PREPARE;
static int32_t ota_e_code = JOY_OTA_ERROR_NONE;
static int8_t  ota_progress = 0;

// 任务取消标志位
static int32_t ota_task_cancel = 0;

/**
 * @brief  ota 获取当前固件版本号
 * 
 * @param[in] version: 输出版本号指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note None.
 *
 */
int32_t device_ota_get_version(char *version)
{
    jl_platform_strcpy(version, ota_version);
    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  ota 获取当前OTA属性
 * 
 * @return 
 * @note None.
 *
 */
JoyProtoKeyValue_t **device_ota_get_properties(int32_t *number)
{
    JoyProtoKeyValue_t **out = NULL;
    int32_t ii;
    int32_t prop_num = 3;
#ifdef D_OTA_PROP_LIST_REQUIRED
    prop_num = 4;
#endif

    out = (JoyProtoKeyValue_t **) jl_platform_malloc(prop_num * sizeof(JoyProtoKeyValue_t *));
    jl_platform_memset(out, 0, prop_num * sizeof(JoyProtoKeyValue_t *));
    for(ii = 0; ii < prop_num; ii++) {
        out[ii] = (JoyProtoKeyValue_t *) jl_platform_malloc(sizeof(JoyProtoKeyValue_t));
        jl_platform_memset(out[ii], 0, sizeof(JoyProtoKeyValue_t));
    }

    // 设备型号
    out[0]->key = (char *) jl_platform_malloc ((jl_platform_strlen(JOYLINK_PROTO_OTA_PROP_MODEL) + 1)* sizeof(char));
    jl_platform_strcpy(out[0]->key, JOYLINK_PROTO_OTA_PROP_MODEL);
    out[0]->value = joylink_proto_keyvalue_pack_string(D_OTA_PROP_MODEL);

    // 是否接收差分包
    out[1]->key = (char *) jl_platform_malloc ((jl_platform_strlen(JOYLINK_PROTO_OTA_PROP_AC_DIFF) + 1)* sizeof(char));
    jl_platform_strcpy(out[1]->key, JOYLINK_PROTO_OTA_PROP_AC_DIFF);
    out[1]->value = joylink_proto_keyvalue_pack_bool(D_OTA_PROP_ACCEPT_DIFF);

    // 签名算法
    out[2]->key = (char *) jl_platform_malloc ((jl_platform_strlen(JOYLINK_PROTO_OTA_PROP_SIG_ALG) + 1)* sizeof(char));
    jl_platform_strcpy(out[2]->key, JOYLINK_PROTO_OTA_PROP_SIG_ALG);
    out[2]->value = joylink_proto_keyvalue_pack_string(D_OTA_PROP_SIG_ALG);

    // ota list
#ifdef D_OTA_PROP_LIST_REQUIRED
    // 签名算法
    out[3]->key = (char *) jl_platform_malloc ((jl_platform_strlen(JOYLINK_PROTO_OTA_PROP_LIST) + 1)* sizeof(char));
    jl_platform_strcpy(out[3]->key, JOYLINK_PROTO_OTA_PROP_LIST);

    out[3]->value = (char *) jl_platform_malloc ((jl_platform_strlen(D_OTA_PROP_OTA_LIST) + 1)* sizeof(char));
    jl_platform_strcpy(out[3]->value, D_OTA_PROP_OTA_LIST);
#endif

    *number = prop_num;

    return out;
}

/**
 * @brief  处理ota方法调用
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] obj_name: 对象空间，可选值{"device", "edge", "group"}
 * @param[in] in_call: 方法调用消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note None.
 *
 */
int32_t device_ota_process_func_call(struct joylink_sdk_t *sdk, char *obj_name, JoyProtoFuncCall_t *in_call)
{
    // 根据方法调用的实例名判断
    if (!jl_platform_strcmp(in_call->functions[0]->key, JOYLINK_PROTO_OTA_FUNC_OTA)) {
        return device_ota_process_func_ota(sdk, obj_name, in_call);
    }
    else if (!jl_platform_strcmp(in_call->functions[0]->key, JOYLINK_PROTO_OTA_FUNC_QUERY_STATE)) {
        return device_ota_process_func_query_state(sdk, obj_name, in_call);
    }
    else if (!jl_platform_strcmp(in_call->functions[0]->key, JOYLINK_PROTO_OTA_FUNC_CANCEL_TASK)) {
        return device_ota_process_func_cancel_task(sdk, obj_name, in_call);
    }
    else {
        log_info("Unsupport OTA Function Call: %s", in_call->functions[0]->key);
    }

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  处理ota方法调用 ota
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] obj_name: 对象空间，可选值{"device", "edge", "group"}
 * @param[in] in_call: 方法调用消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note None.
 *
 */
int32_t device_ota_process_func_ota(struct joylink_sdk_t *sdk, char *obj_name, JoyProtoFuncCall_t *in_call)
{
    log_info("OTA Function Key:%s", in_call->functions[0]->key);
    // 方法响应
    JoyProtoFuncCallRes_t call_res;
    jl_platform_memset(&call_res, 0, sizeof(JoyProtoFuncCallRes_t));
    call_res.deviceId = in_call->deviceId;
    call_res.messageId = in_call->messageId;
    call_res.code = 200;

    // OTA 参数处理
    int32_t ii;
    char *tmp_char = NULL;
    double tmp_double = 0.0;

    // TODO 必要的初始化

    for (ii = 0; ii < in_call->functions[0]->in_num; ii++) {
        jl_platform_printf("\t%s:\t\t%s\r\n", in_call->functions[0]->in[ii]->key, in_call->functions[0]->in[ii]->value);

        // 暂时保存需要的信息
        if (!jl_platform_strcmp(in_call->functions[0]->in[ii]->key, D_OTA_FUNC_ARG_URL)) {
            tmp_char = joylink_proto_keyvalue_parse_string(in_call->functions[0]->in[ii]->value);
            if (tmp_char != NULL) {
                jl_platform_strcpy(ota_url, tmp_char);
                jl_platform_free(tmp_char);
                tmp_char = NULL;
            }
        }
        // 判断安装方式
        else if (!jl_platform_strcmp(in_call->functions[0]->in[ii]->key, D_OTA_FUNC_ARG_INSTALL_MODE)) {
            tmp_char = joylink_proto_keyvalue_parse_string(in_call->functions[0]->in[ii]->value);
            if (tmp_char != NULL) {
                jl_platform_strcpy(ota_install_model, tmp_char);
                jl_platform_free(tmp_char);
                tmp_char = NULL;
            }
        }
        // 保存任务标识
        else if (!jl_platform_strcmp(in_call->functions[0]->in[ii]->key, D_OTA_FUNC_ARG_TASK_ID)) {
            tmp_char = joylink_proto_keyvalue_parse_string(in_call->functions[0]->in[ii]->value);
            if (tmp_char != NULL) {
                jl_platform_strcpy(ota_task_id, tmp_char);
                jl_platform_free(tmp_char);
                tmp_char = NULL;
            }
        }
        // 版本号
        else if (!jl_platform_strcmp(in_call->functions[0]->in[ii]->key, D_OTA_FUNC_ARG_VERSION)) {
            tmp_char = joylink_proto_keyvalue_parse_string(in_call->functions[0]->in[ii]->value);
            if (tmp_char != NULL) {
                jl_platform_strcpy(ota_task_version, tmp_char);
                jl_platform_free(tmp_char);
                tmp_char = NULL;
            }
        }
        // 升级包名称
        else if (!jl_platform_strcmp(in_call->functions[0]->in[ii]->key, D_OTA_FUNC_ARG_PKG_NAME)) {
            tmp_char = joylink_proto_keyvalue_parse_string(in_call->functions[0]->in[ii]->value);
            if (tmp_char != NULL) {
                jl_platform_strcpy(ota_package_name, tmp_char);
                jl_platform_free(tmp_char);
                tmp_char = NULL;
            }
        }
        // 升级时间
        else if (!jl_platform_strcmp(in_call->functions[0]->in[ii]->key, D_OTA_FUNC_ARG_TIME)) {
            joylink_proto_keyvalue_parse_double(in_call->functions[0]->in[ii]->value, &tmp_double);
            ota_task_exec_time = (uint32_t)(tmp_double/1000);
        }
        else {}
    }


    // TODO 为了防止降级，请确认版本号
    /*
    if(ota_task_version < ota_version) {
        call_res.code = 204;
    }
    */

    // TODO OTA 任务线程
    // 需要根据实际的升级类型选择选择执行逻辑
    jl_thread_t task_id;
    if (call_res.code == 200) {
        task_id.thread_task = (threadtask) device_ota_task_thread;
	    task_id.stackSize = 0x4000;
        task_id.priority = JL_THREAD_PRI_DEFAULT;
        task_id.parameter = (void *) sdk;
        jl_platform_thread_start(&task_id);
    }

    // 方法调用响应
    call_res.timestamp = jl_platform_time(NULL);
    call_res.func_num = 1;
    call_res.functions = (JoyProtoFuncCallResFunc_t **) jl_platform_malloc(
                      call_res.func_num * sizeof(JoyProtoFuncCallResFunc_t *));
    jl_platform_memset(call_res.functions, 0, call_res.func_num * sizeof(JoyProtoFuncCallResFunc_t *));
    call_res.functions[0] = joylink_proto_ota_create_func_res(in_call->functions[0]->key);
    
    joylink_sdk_dev_func_call_response(sdk, obj_name, JOYLINK_PROTO_SERVICE_OTA, &call_res);

    // 释放内存
    call_res.deviceId = NULL;
    call_res.messageId = NULL;
    joylink_proto_free_func_call_res(&call_res);
    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  处理ota方法调用 查询状态
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] obj_name: 对象空间，可选值{"device", "edge", "group"}
 * @param[in] in_call: 方法调用消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note None.
 *
 */
int32_t device_ota_process_func_query_state(struct joylink_sdk_t *sdk, char *obj_name, JoyProtoFuncCall_t *in_call)
{
    log_info("OTA Function Key:%s", in_call->functions[0]->key);
    int32_t ii;
    char *tmp_string = NULL;
    for (ii = 0; ii < in_call->functions[0]->in_num; ii++) {
        jl_platform_printf("\t%s:\t\t%s\r\n", in_call->functions[0]->in[ii]->key, in_call->functions[0]->in[ii]->value);

        // 获取任务标识
        if (!jl_platform_strcmp(in_call->functions[0]->in[ii]->key, D_OTA_FUNC_ARG_TASK_ID)) {
            tmp_string = joylink_proto_keyvalue_parse_string(in_call->functions[0]->in[ii]->value);
        }
    }

    // TODO
    // 执行状态查询

    // 方法响应
    JoyProtoFuncCallRes_t call_res;
    jl_platform_memset(&call_res, 0, sizeof(JoyProtoFuncCallRes_t));
    call_res.deviceId = in_call->deviceId;
    call_res.messageId = in_call->messageId;
    call_res.timestamp = jl_platform_time(NULL);
    call_res.code = 200;
    call_res.func_num = 1;
    call_res.functions = (JoyProtoFuncCallResFunc_t **) jl_platform_malloc(
                      call_res.func_num * sizeof(JoyProtoFuncCallResFunc_t *));
    jl_platform_memset(call_res.functions, 0, call_res.func_num * sizeof(JoyProtoFuncCallResFunc_t *));
    call_res.functions[0] = joylink_proto_ota_create_func_query_res(in_call->functions[0]->key,
                                                                    tmp_string, ota_state, ota_e_code, ota_progress);

    joylink_sdk_dev_func_call_response(sdk, obj_name, JOYLINK_PROTO_SERVICE_OTA, &call_res);

    // 释放内存
    call_res.deviceId = NULL;
    call_res.messageId = NULL;
    joylink_proto_free_func_call_res(&call_res);

    if (tmp_string != NULL) {
        jl_platform_free(tmp_string);
    }

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  处理ota方法调用 取消任务
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] obj_name: 对象空间，可选值{"device", "edge", "group"}
 * @param[in] in_call: 方法调用消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note None.
 *
 */
int32_t device_ota_process_func_cancel_task(struct joylink_sdk_t *sdk, char *obj_name, JoyProtoFuncCall_t *in_call)
{
    log_info("OTA Function Key:%s", in_call->functions[0]->key);
    int32_t ii;
    for (ii = 0; ii < in_call->functions[0]->in_num; ii++) {
        jl_platform_printf("\t%s:\t\t%s\r\n", in_call->functions[0]->in[ii]->key, in_call->functions[0]->in[ii]->value);
    }

    // TODO
    // 取消OTA
    ota_task_cancel = 1;

    // 方法响应
    JoyProtoFuncCallRes_t call_res;
    jl_platform_memset(&call_res, 0, sizeof(JoyProtoFuncCallRes_t));
    call_res.deviceId = in_call->deviceId;
    call_res.messageId = in_call->messageId;
    call_res.timestamp = jl_platform_time(NULL);
    call_res.code = 200;
    call_res.func_num = 1;
    call_res.functions = (JoyProtoFuncCallResFunc_t **) jl_platform_malloc(
                      call_res.func_num * sizeof(JoyProtoFuncCallResFunc_t *));
    jl_platform_memset(call_res.functions, 0, call_res.func_num * sizeof(JoyProtoFuncCallResFunc_t *));
    call_res.functions[0] = joylink_proto_ota_create_func_res(in_call->functions[0]->key);

    joylink_sdk_dev_func_call_response(sdk, obj_name, JOYLINK_PROTO_SERVICE_OTA, &call_res);

    // 释放内存
    call_res.deviceId = NULL;
    call_res.messageId = NULL;
    joylink_proto_free_func_call_res(&call_res);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  检测任务是否取消
 * 
 */
static int32_t ota_task_step_check_cancel(struct joylink_sdk_t *sdk) {
    if (ota_task_cancel == 1) {
        ota_state = JOY_OTA_ST_FAILURE;
        ota_e_code = JOY_OTA_ERROR_TASK_CL;
        joylink_sdk_dev_ota_state_changed_post(sdk, "device", ota_task_id, ota_state, ota_e_code, ota_progress);
        return 1;
    }
    return 0;
}

/**
 * @brief  执行升级包下载任务
 * 
 */

static void ota_package_download_procedure(void) {
    /*
    char cmd_string[512];
    jl_platform_memset(cmd_string, 0, 512 * sizeof(char));
    jl_platform_sprintf(cmd_string, "%s %s", "wget", ota_url);
    log_info("Execute Command: %s", cmd_string);
    system(cmd_string);
    jl_platform_msleep(3000);
    */
}

/**
 * @brief  执行正常安装流程
 * 
 */

static void ota_install_procedure(struct joylink_sdk_t *sdk) 
{
    // TODO 必要的文件校验
    // 如果文件校验有错误 上报状态 并退出
    // joylink_sdk_dev_ota_state_changed_post(sdk, "device", ota_task_id, JOY_OTA_ST_FAILURE, JOY_OTA_ERROR_SIG_VER_FAIL, 10);
    // return;

    // 上报状态变化 正在安装
    jl_platform_msleep(5000);
    if (ota_task_step_check_cancel(sdk)) {
        return;
    }
    ota_state = JOY_OTA_ST_INSTALLING;
    ota_progress = 50;
    joylink_sdk_dev_ota_state_changed_post(sdk, "device", ota_task_id, ota_state, ota_e_code, ota_progress);

    // 上报状态变化 升级完成
    jl_platform_msleep(5000);
    if (ota_task_step_check_cancel(sdk)) {
        return;
    }
    ota_state = JOY_OTA_ST_INSTALL_DONE;
    ota_progress = 100;
    joylink_sdk_dev_ota_state_changed_post(sdk, "device", ota_task_id, JOY_OTA_ST_INSTALL_DONE, JOY_OTA_ERROR_NONE, 100);

    // 上报 当前版本号
    jl_platform_msleep(5000);
    jl_platform_strcpy(ota_version , ota_task_version); // 更新版本号
    joylink_sdk_dev_ota_version_changed_post(sdk, "device", NULL, ota_version);
} 

/**
 * @brief  设备ota模拟任务线程
 * 
 * @note 以下只是流程示例，需要根据实际情况进行修改.
 *
 */
void device_ota_task_thread(void *in_param)
{
    log_info("Enter device_ota_task_thread");

    // 局部变量 当前时间
    uint32_t current_time = 0;

    // 初始化任务状态
    ota_task_cancel = 0;
    ota_state = JOY_OTA_ST_PREPARE;
    ota_e_code = JOY_OTA_ERROR_NONE;
    ota_progress = 0;

    struct joylink_sdk_t *sdk = (struct joylink_sdk_t *)in_param;
    jl_platform_msleep(5000);

    // 上报状态变化 正在准备
    ota_state = JOY_OTA_ST_PREPARE;
    joylink_sdk_dev_ota_state_changed_post(sdk, "device", ota_task_id, ota_state, ota_e_code, ota_progress);

    // TODO 如果安装方式为 用户确认后下载安装 需要增加与用户的的交互逻辑
    if (!jl_platform_strcmp(ota_install_model, D_OTA_INSTALL_MODE_C_D_I)) {
        ota_state = JOY_OTA_ST_CF_DOWNLOAD;
        joylink_sdk_dev_ota_state_changed_post(sdk, "device", ota_task_id, ota_state, ota_e_code, ota_progress);
#if 0
    // TODO 交互逻辑 如果用户取消 上报状态后退出

#endif
    }

    // 上报状态变化 正在下载
    jl_platform_msleep(5000);
    if (ota_task_step_check_cancel(sdk)) {
        return;
    }
    ota_state = JOY_OTA_ST_DOWNLOADIND;
    ota_progress = 1;
    joylink_sdk_dev_ota_state_changed_post(sdk, "device", ota_task_id, ota_state, ota_e_code, ota_progress);

    // 下载固件文件
    ota_package_download_procedure();

    // 上报状态变化 下载完成
    jl_platform_msleep(5000);
    if (ota_task_step_check_cancel(sdk)) {
        return;
    }
    ota_state = JOY_OTA_ST_DOWNLOAD_DONE;
    ota_progress = 10;
    joylink_sdk_dev_ota_state_changed_post(sdk, "device", ota_task_id, ota_state, ota_e_code, ota_progress);

    // 根据不同的安装模式 进行处理
    // 等待客户确认安装
    if (!jl_platform_strcmp(ota_install_model, D_OTA_INSTALL_MODE_C_D_I) ||
        !jl_platform_strcmp(ota_install_model, D_OTA_INSTALL_MODE_C_I)) {
        ota_state = JOY_OTA_ST_CF_INSTALL;
        joylink_sdk_dev_ota_state_changed_post(sdk, "device", ota_task_id, ota_state, ota_e_code, ota_progress);
#if 0
        // TODO 交互逻辑
#endif
    }
    // 定时安装
    else if (!jl_platform_strcmp(ota_install_model, D_OTA_INSTALL_MODE_TIMING)) {
        // TODO 增加定时器 来执行
        while (1) {
            if (ota_task_cancel) {
                ota_state = JOY_OTA_ST_FAILURE;
                ota_e_code = JOY_OTA_ERROR_DOWNLOAD_CL;
                joylink_sdk_dev_ota_state_changed_post(sdk, "device", ota_task_id, ota_state, ota_e_code, ota_progress);
                return;
            }
            // 判断是否到时间
            current_time = jl_platform_time(NULL);
            if (current_time >= ota_task_exec_time) {
                log_info("Current time [%d] is Enquel Setted Time [%d], Starting to Update", current_time, ota_task_exec_time);
                break;
            }

            jl_platform_msleep(200);
        }
    }
    // 闲时安装
    else if (!jl_platform_strcmp(ota_install_model, D_OTA_INSTALL_MODE_TIMING)) {
        // TODO 增加逻辑 判断系统是否是空闲状态
    } else {};


    // 执行正常升级流程
    ota_install_procedure(sdk);

    log_info("Leave device_ota_task_thread");
}

// end of file