/*Copyright (c) 2015-2050, JD All rights reserved.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License. */

/* --------------------------------------------------
 * @brief: 源文件 一个设备示例
 *
 * @version: 1.0
 *
 * --------------------------------------------------
 */

#include <stdio.h>
#include "joylink_sub_device.h"
#include "joylink_string.h"
#include "joylink_memory.h"
#include "joylink_log.h"
#include "joylink_time.h"
#include "joylink_thread.h"

#include "joylink_protocol.h"
#include "joylink_sdk.h"

#include "cJSON.h"
#include "joylink_device_ota.h"

// 示例属性实例名
#define PROP_KEY        ("light.power")

// 示例事件实例名、事件参数实例名
#define EVENT_KEY       ("light.alert")
#define EVENT_ARG_KEY   ("dateAlert")

// 定义子设备信息,这里暂时只代理一个子设备
#define SUB_DEVICE_ID       ("001")

// 这只是一个示例，开发者可根据实际的硬件资源来确定维护方式
// 这里仅仅是为了展示如何使用SDK API
typedef struct {
    JoyProtoSubDevice_t info;
    uint16_t  prop_version;
    bool light_power; // properties
} sub_device_t;
static sub_device_t sub_device;

/**
 * @brief  子设备初始化
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 */
int32_t sub_device_initialise(struct joylink_sdk_t *sdk)
{
    jl_platform_memset(&sub_device, 0, sizeof(sub_device_t));
    sub_device.info.name = "sub_light";
    sub_device.info.deviceId = SUB_DEVICE_ID;

    sub_device.prop_version = 0;
    sub_device.light_power = true;

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  属性设置 回调函数
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] in_set: 属性设置消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 *    sdk 收到属性设置消息时 会调用该函数
 */
int32_t sub_device_callback_property_set(struct joylink_sdk_t *sdk, JoyProtoPropSet_t *in_set)
{
    // 发送 属性设置响应示例代码
    JoyProtoPropSetRes_t out_res;
    jl_platform_memset(&out_res, 0, sizeof(JoyProtoPropSetRes_t));
    out_res.deviceId = in_set->deviceId;
    out_res.messageId = in_set->messageId;
    out_res.timestamp = jl_platform_time(NULL);
    
    // 检查请求版本号
    if (joylink_sdk_prop_version_check(sub_device.prop_version, in_set->version)) {
        out_res.code = 200;
    }
    else {
        out_res.code = 204;
    }
    
    if (out_res.code == 200) {
        // 检查属性实例名
        if (!jl_platform_strcmp(in_set->properties[0]->key, PROP_KEY)) {
            // 解析期望的值
            joylink_proto_keyvalue_parse_bool(in_set->properties[0]->value, &sub_device.light_power);
            
            // 更新版本号
            sub_device.prop_version = in_set->version;
            
            // 给出响应
            out_res.prop_num = 1;
            out_res.properties = (JoyProtoKeyValue_t **) jl_platform_malloc(out_res.prop_num * sizeof(JoyProtoKeyValue_t *));
            jl_platform_memset(out_res.properties, 0, out_res.prop_num * sizeof(JoyProtoKeyValue_t *));
            out_res.properties[0] = (JoyProtoKeyValue_t *) jl_platform_malloc(sizeof(JoyProtoKeyValue_t));
            out_res.properties[0]->key = (char *) jl_platform_malloc (32 * sizeof(char));
            jl_platform_strcpy(out_res.properties[0]->key, PROP_KEY);
            out_res.properties[0]->value = joylink_proto_keyvalue_pack_bool(sub_device.light_power);
            out_res.version = sub_device.prop_version;
        }
        else {
            out_res.code = 400;
        }
    }

    joylink_sdk_sub_dev_prop_set_response(sdk, &out_res);
    
    out_res.deviceId = NULL;
    out_res.messageId = NULL;
    joylink_proto_free_prop_set_res(&out_res);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  属性获取 回调函数
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] in_get: 属性获取消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 *    sdk 收到属性设置消息时 会调用该函数
 */
int32_t sub_device_callback_property_get(struct joylink_sdk_t *sdk, JoyProtoPropGet_t *in_get)
{
    // 发送 属性设置响应示例代码
    JoyProtoPropGetRes_t out_res;
    jl_platform_memset(&out_res, 0, sizeof(JoyProtoPropGetRes_t));
    out_res.deviceId = in_get->deviceId;
    out_res.messageId = in_get->messageId;
    out_res.timestamp = jl_platform_time(NULL);
    /*
     * out_res.properties[0]->value = joylink_proto_keyvalue_pack_int32(200);
     * 注意内存释放问题
     */
    out_res.prop_num = 1;
    out_res.properties = (JoyProtoKeyValue_t **) jl_platform_malloc(out_res.prop_num * sizeof(JoyProtoKeyValue_t *));
    jl_platform_memset(out_res.properties, 0, out_res.prop_num * sizeof(JoyProtoKeyValue_t *));
    out_res.properties[0] = (JoyProtoKeyValue_t *) jl_platform_malloc(sizeof(JoyProtoKeyValue_t));
    out_res.properties[0]->key = (char *) jl_platform_malloc (32 * sizeof(char));
    jl_platform_strcpy(out_res.properties[0]->key, PROP_KEY);
    out_res.properties[0]->value = joylink_proto_keyvalue_pack_bool(sub_device.light_power);
    
    out_res.version = sub_device.prop_version;
    out_res.code = 200;

    joylink_sdk_sub_dev_prop_get_response(sdk, &out_res);
    
    out_res.deviceId = NULL;
    out_res.messageId = NULL;
    joylink_proto_free_prop_get_res(&out_res);
    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  方法调用 回调函数
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] in_call: 方法调用消息结构体指针
 * @return 
 *    返回值 JoylinkRetCode_E
 * @note
 *    sdk 收到方法调用消息时 会调用该函数
 */
int32_t sub_device_callback_function_call(struct joylink_sdk_t *sdk, JoyProtoFuncCall_t *in_call)
{
    log_error("Enter sub_device_callback_function_call");
    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  获取子设备信息
 * 
 * @param[in] subdeviceId: 子设备设备标识
 * @return 
 *    返回值 JoylinkRetCode_E
 */
JoyProtoSubDevice_t *sub_device_get_info(char *subdeviceId)
{
    return &(sub_device.info);
}

/**
 * @brief  子设备属性上报
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] subdeviceId: 子设备设备标识
 * @return 
 *    返回值 JoylinkRetCode_E
 */
int32_t sub_device_properties_post(struct joylink_sdk_t *sdk, char *subdeviceId)
{
    JoyProtoPropPost_t prop_post;
    jl_platform_memset(&prop_post, 0, sizeof(JoyProtoPropPost_t));
    prop_post.deviceId = sub_device.info.deviceId;
    prop_post.timestamp = jl_platform_time(NULL);
    prop_post.version = sub_device.prop_version;
    prop_post.messageId = joylink_sdk_get_messageId(NULL);

    prop_post.prop_num = 1;
    prop_post.properties = (JoyProtoKeyValue_t **) jl_platform_malloc(prop_post.prop_num * sizeof(JoyProtoKeyValue_t *));
    jl_platform_memset(prop_post.properties, 0, prop_post.prop_num * sizeof(JoyProtoKeyValue_t *));
    prop_post.properties[0] = (JoyProtoKeyValue_t *) jl_platform_malloc(prop_post.prop_num * sizeof(JoyProtoKeyValue_t));
    prop_post.properties[0]->key = (char *) jl_platform_malloc(32 * sizeof(char));
    jl_platform_strcpy(prop_post.properties[0]->key, PROP_KEY);
    prop_post.properties[0]->value = joylink_proto_keyvalue_pack_bool(sub_device.light_power);

    joylink_sdk_sub_dev_prop_post(sdk, &prop_post);
    prop_post.deviceId = NULL;
    joylink_proto_free_prop_post(&prop_post);

    return JOYLINK_RET_SUCCESS;
}

/**
 * @brief  子设备事件上报
 * 
 * @param[in] sdk: joylink sdk 结构体指针
 * @param[in] subdeviceId: 子设备设备标识
 * @return 
 *    返回值 JoylinkRetCode_E
 */
int32_t sub_device_event_post(struct joylink_sdk_t *sdk, char *subdeviceId)
{
    jl_time_stamp_t time_stamp;
    char date[32];

    JoyProtoEvtPost_t evt_post;
    jl_platform_memset(&evt_post, 0, sizeof(JoyProtoEvtPost_t));
    evt_post.deviceId = sub_device.info.deviceId;
    evt_post.timestamp = jl_platform_time(NULL);
    evt_post.messageId = joylink_sdk_get_messageId(NULL);
    evt_post.evt_num = 1;
    evt_post.events = (JoyProtoEvtPostEvt_t **) jl_platform_malloc (evt_post.evt_num * sizeof(JoyProtoEvtPostEvt_t *));
    evt_post.events[0] = (JoyProtoEvtPostEvt_t *) jl_platform_malloc (sizeof(JoyProtoEvtPostEvt_t));
    jl_platform_memset(evt_post.events[0], 0, sizeof(JoyProtoEvtPostEvt_t));
    evt_post.events[0]->key = (char *) jl_platform_malloc(32 * sizeof(char));
    jl_platform_strcpy(evt_post.events[0]->key, EVENT_KEY);
    evt_post.events[0]->param_num = 1;
    evt_post.events[0]->parameters = (JoyProtoKeyValue_t **) jl_platform_malloc(evt_post.events[0]->param_num * sizeof(JoyProtoKeyValue_t *));
    evt_post.events[0]->parameters[0] = (JoyProtoKeyValue_t *) jl_platform_malloc(sizeof(JoyProtoKeyValue_t));
    jl_platform_memset(evt_post.events[0]->parameters[0], 0, sizeof(JoyProtoKeyValue_t));
    evt_post.events[0]->parameters[0]->key = (char *) jl_platform_malloc(32 * sizeof(char));
    jl_platform_strcpy(evt_post.events[0]->parameters[0]->key, EVENT_ARG_KEY);

    jl_platform_time_get_timestamp(&time_stamp);
    jl_platform_sprintf(date, "%d%d", time_stamp.second, time_stamp.ms);
    evt_post.events[0]->parameters[0]->value = joylink_proto_keyvalue_pack_string(date);

    // 上报事件
    joylink_sdk_sub_dev_evt_post(sdk, &evt_post);

    // 释放资源
    evt_post.deviceId = NULL;
    joylink_proto_free_evt_post(&evt_post);
}

// end of file
