#include "joylink_stdio.h"
#include "joylink_mqtt.h"
#include "joylink_stdint.h"
#include "joylink_log.h"
#include "joylink_thread.h"

#define __MQTT_MOSQUITTO_PAL__

#ifdef __MQTT_MOSQUITTO_PAL__
#include "mosquitto.h"
#include "mosquitto_internal.h"
#include "mqtt_protocol.h"

// static variable of mosquitto handler
static struct mosquitto *my_mosq = NULL;

// call back struct defination
struct mosq_call_back_t {
    int32_t (*on_connect)(jl_mqtt_t, void *);
    void * on_connect_user_data;
    int32_t (*on_disconnect)(jl_mqtt_t, void *);
    void * on_disconnect_user_data;
    int32_t (*on_message)(jl_mqtt_t, jl_mqtt_msg_t *, void *);
    void * on_message_user_data;
    int32_t (*on_subscribe)(jl_mqtt_t, jl_mqtt_msg_t *, void *);
    void * on_subscribe_user_data;
    int32_t (*on_publish)(jl_mqtt_t, jl_mqtt_msg_t *, void *);
    void * on_publish_user_data;
};

// static variable of call back struct pointer
static struct mosq_call_back_t mosq_cb = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};


// call back on mosquitto connect 
static void mosq_connect_callback(struct mosquitto *mosq, void *obj, int result, int flags, const mosquitto_property *properties)
{
    log_info("Enter mosq_connect_callback");
    if (mosq_cb.on_connect != NULL) {
        mosq_cb.on_connect((jl_mqtt_t) mosq, mosq_cb.on_connect_user_data);
    }
    log_info("Leave mosq_connect_callback");
}

// call back on mosquitto disconnect
static mosq_disconnect_callback(struct mosquitto *mosq, void *obj, int flags, const mosquitto_property *properties)
{
    log_info("Enter mosq_disconnect_callback");
    if (mosq_cb.on_disconnect != NULL) {
        mosq_cb.on_disconnect((jl_mqtt_t) mosq, mosq_cb.on_disconnect_user_data);
    }
    log_info("Leave mosq_disconnect_callback");
}

// call back on mosquitto message recieved
static void mosq_message_callback(struct mosquitto *mosq, void *obj, const struct mosquitto_message *message, const mosquitto_property *properties)
{
    if (mosq_cb.on_message != NULL) {
        jl_mqtt_msg_t joy_msg;
        memset(&joy_msg, 0, sizeof(jl_mqtt_msg_t));
        joy_msg.topic = message->topic;
        joy_msg.payload = message->payload;
        joy_msg.qos = message->qos;
        joy_msg.mid = message->mid;
        
        mosq_cb.on_message((jl_mqtt_t)mosq, &joy_msg, mosq_cb.on_message_user_data);
    }
}

// call back on mosquitto log print
void mosq_log_callback(struct mosquitto *mosq, void *obj, int level, const char *str)
{
	UNUSED(mosq);
	UNUSED(obj);
	UNUSED(level);

	printf("%s\n", str);
}

// call back on subscribed
void mosq_subscribe_callback(struct mosquitto *mosq, void *obj, int mid, int qos_count, const int *granted_qos)
{
    if (mosq_cb.on_subscribe != NULL) {
        mosq_cb.on_subscribe((jl_mqtt_t)mosq, mid, mosq_cb.on_subscribe_user_data);
    }
}

// call back on publish
void mosq_publish_callback(struct mosquitto *mosq, void *obj, int mid, int reason_code, const mosquitto_property *properties)
{
    if (mosq_cb.on_publish != NULL) {
        mosq_cb.on_publish((jl_mqtt_t)mosq, mid, mosq_cb.on_publish_user_data);
    }
}

// static variable definitial for loop yeild
static int run_once = 0;
static int run = 1;
static unsigned int reconnects = 0;
static unsigned long reconnect_delay;
static struct timeval delay;

// static function of loop yeild
static int client_task_loop_yield(struct mosquitto *mosq, int timeout, int max_packets)
{
	int rc;

	if(!mosq) return MOSQ_ERR_INVAL;
    if (run_once == 0) {
	    if(mosq->state == mosq_cs_connect_async){
	    	mosquitto_reconnect(mosq);
	    }
        run_once = 1;
    }

	// while(run){
		do{
			rc = mosquitto_loop(mosq, timeout, max_packets);
			if (reconnects !=0 && rc == MOSQ_ERR_SUCCESS){
				reconnects = 0;
			}
            
            //FIXME return for one step
            if (rc == MOSQ_ERR_SUCCESS) {
                return rc;
            }

		}while(run && rc == MOSQ_ERR_SUCCESS);
		/* Quit after fatal errors. */
		switch(rc){
			case MOSQ_ERR_NOMEM:
			case MOSQ_ERR_PROTOCOL:
			case MOSQ_ERR_INVAL:
			case MOSQ_ERR_NOT_FOUND:
			case MOSQ_ERR_TLS:
			case MOSQ_ERR_PAYLOAD_SIZE:
			case MOSQ_ERR_NOT_SUPPORTED:
			case MOSQ_ERR_AUTH:
			case MOSQ_ERR_ACL_DENIED:
			case MOSQ_ERR_UNKNOWN:
			case MOSQ_ERR_EAI:
			case MOSQ_ERR_PROXY:
                log_error("Mosquitto Encounter Fatal Error:%d\n", rc);
				return rc;
			case MOSQ_ERR_ERRNO:
                log_error("Mosquitto Encounter MOSQ_ERR_ERRNO:%d, Error Number:%d\n", rc, errno);
				break;
		}
		if(errno == EPROTO){
            log_error("Mosquitto Quit Cause of Error Number: EPROTO\n");
			return rc;
		}
		do{
			rc = MOSQ_ERR_SUCCESS;
			pthread_mutex_lock(&mosq->state_mutex);
			if(mosq->state == mosq_cs_disconnecting){
				run = 0;
				pthread_mutex_unlock(&mosq->state_mutex);
			}else{
				pthread_mutex_unlock(&mosq->state_mutex);

				if(mosq->reconnect_delay_max > mosq->reconnect_delay){
					if(mosq->reconnect_exponential_backoff){
						reconnect_delay = mosq->reconnect_delay*(reconnects+1)*(reconnects+1);
					}else{
						reconnect_delay = mosq->reconnect_delay*(reconnects+1);
					}
				}else{
					reconnect_delay = mosq->reconnect_delay;
				}

				if(reconnect_delay > mosq->reconnect_delay_max){
					reconnect_delay = mosq->reconnect_delay_max;
				}else{
					reconnects++;
				}
 
                // reconnect
                log_info("MQTT Reconnect 5 Seconds Later, Time Now %d", jl_platform_time(NULL));
                jl_platform_msleep(5000);
/*
				req.tv_sec = reconnect_delay;
				req.tv_nsec = 0;
				while(nanosleep(&req, &rem) == -1 && errno == EINTR){
					req = rem;
				}
*/
				delay.tv_sec = reconnect_delay;
				delay.tv_usec = 0;
				select(0, NULL, NULL, NULL, &delay);

				pthread_mutex_lock(&mosq->state_mutex);
				if(mosq->state == mosq_cs_disconnecting){
					run = 0;
					pthread_mutex_unlock(&mosq->state_mutex);
				}else{
					pthread_mutex_unlock(&mosq->state_mutex);
					rc = mosquitto_reconnect(mosq);
				}
			}
		}while(run && rc != MOSQ_ERR_SUCCESS);
	//}
	return rc;
}

#endif // __MQTT_MOSQUITTO_PAL__

/**
 * @brief   创建一个MQTT客户端
 *
 * @param[in] config: MQTT的设置参指针
 * @return 
 	MQTT客户端句柄
 * @see None.
 * @note None.
 */
jl_mqtt_t jl_platform_mqtt_create(jl_mqtt_config_t *config)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    // mosquitto libarary initialise
    mosquitto_lib_init();

    // create mosquitto
    my_mosq = mosquitto_new(config->clientId, true, NULL);
    if(!my_mosq){
		return NULL;
	}
    
    // set options mqtt versions
    mosquitto_int_option(my_mosq, MOSQ_OPT_PROTOCOL_VERSION, MQTT_PROTOCOL_V311);

    // set options username and password
    if ((config->username != NULL) || (config->password != NULL)) {
        mosquitto_username_pw_set(my_mosq, config->username, config->password);
    }

    // set options tls
    if (!strcmp(config->protocol, "mqtt_tls")) {
        mosquitto_tls_set(my_mosq, config->cafile, NULL, config->cert, config->key, NULL);
    }

    // set options insecure
    if (config->insecure) {
        mosquitto_tls_insecure_set(my_mosq, true);
    }

    // set options max message inflight
    mosquitto_max_inflight_messages_set(my_mosq, 20);

    mosquitto_log_callback_set(my_mosq, mosq_log_callback);
	mosquitto_connect_v5_callback_set(my_mosq, mosq_connect_callback);
    mosquitto_disconnect_v5_callback_set(my_mosq, mosq_disconnect_callback);
	mosquitto_message_v5_callback_set(my_mosq, mosq_message_callback);

    mosquitto_subscribe_v5_callback_set(my_mosq, mosq_subscribe_callback);
    mosquitto_publish_v5_callback_set(my_mosq, mosq_publish_callback);
    
    return (jl_mqtt_t) my_mosq;
#else
    return NULL;
#endif
}

/**
 * @brief   销毁一个MQTT客户端
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_destory(jl_mqtt_t mqtt_handler)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    mosquitto_lib_cleanup();

	if (mqtt_handler) {
		mosquitto_destroy((struct mosquitto *)mqtt_handler);
        my_mosq = NULL;
        mqtt_handler = NULL;
	}
    return 0;
#else
    return -1;
#endif
}

/**
 * @brief   MQTT客户端连接到broker
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @param[in] cfg: MQTT 配置参数指针
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_connect(jl_mqtt_t mqtt_handler, jl_mqtt_config_t *cfg)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    struct mosquitto *mosq = (struct mosquitto *)mqtt_handler;

    int rc;
    rc = mosquitto_connect_bind_v5(mosq, cfg->hostname, cfg->port, 60, NULL, NULL);
    if(rc>0){
        if(rc == MOSQ_ERR_ERRNO){
            printf("Error: mosquitto_connect_bind_v5:%d, %d\n", MOSQ_ERR_ERRNO, errno);
        } else {
            printf(cfg, "Unable to connect (%s).\n", mosquitto_strerror(rc));
        }
        mosquitto_lib_cleanup();
        return rc;
    }
    return 0;
#else
    return -1;
#endif
}

/**
 * @brief   MQTT客户端断开连接
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_disconnect(jl_mqtt_t mqtt_handler)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    struct mosquitto *mosq = (struct mosquitto *)mqtt_handler;
    return mosquitto_disconnect(mosq);
#else
    return -1;
#endif
}

/**
 * @brief   MQTT客户端运单步行执行函数，会循环调用
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @param[in] timeout_ms: 运行超时时间，单位：毫秒
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_yield(jl_mqtt_t mqtt_handler, int32_t timeout_ms)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    struct mosquitto *mosq = (struct mosquitto *)mqtt_handler;
    return client_task_loop_yield(mosq, -1, 1);
#else
    return -1;
#endif
}

/**
 * @brief   MQTT客户端 订阅消息主题函数
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @param[in] topic: 订阅主题字符串的指针
 * @param[in] qos: 对应主题的消息服务等级
 * @param[in] mid: 对应MQTT的消息id
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_subscribe(jl_mqtt_t mqtt_handler, char *topic, int32_t qos, int32_t *mid)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    struct mosquitto *mosq = (struct mosquitto *)mqtt_handler;
    return mosquitto_subscribe(mosq, mid, topic, qos);
#else
    return -1;
#endif
}

/**
 * @brief   MQTT客户端 消息发布函数
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @param[in] message: MQTT 消息结构体指针
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_publish(jl_mqtt_t mqtt_handler, jl_mqtt_msg_t *message)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    struct mosquitto *mosq = (struct mosquitto *)mqtt_handler;
    mosquitto_publish(mosq, &message->mid,
                        message->topic, 
                        strlen(message->payload), 
                        message->payload,
                        message->qos,
                        true);
    return 0;
#else
    return -1;
#endif
}

/**
 * @brief   MQTT客户端 设置 连接成功回调函数
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @param[in] cb_connect: MQTT 连接成功后需要调用的函数句柄
 * @param[in] user_data: 用户数据
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_set_cb_connect(jl_mqtt_t mqtt_handler, int32_t (*cb_connect)(jl_mqtt_t ,void *), void *user_data)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    mosq_cb.on_connect = cb_connect;
    mosq_cb.on_connect_user_data = user_data;
    return 0;
#else
    return -1;
#endif
}

/**
 * @brief   MQTT客户端 设置 连接断开回调函数
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @param[in] cb_disconnect: MQTT 连接断开后需要调用的函数句柄
 * @param[in] user_data: 用户数据
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_set_cb_disconnect(jl_mqtt_t mqtt_handler, int32_t (*cb_disconnect)(jl_mqtt_t, void *), void *user_data)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    mosq_cb.on_disconnect = cb_disconnect;
    mosq_cb.on_disconnect_user_data = user_data;
    return 0;
#else
    return -1;
#endif    
}

/**
 * @brief   MQTT客户端 设置 收到消息后的回调函数
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @param[in] cb_message: MQTT 收到消息后需要调用的回调函数
 * @param[in] user_data: 用户数据
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_set_cb_message(jl_mqtt_t mqtt_handler, int32_t (*cb_message)(jl_mqtt_t, jl_mqtt_msg_t *, void *), void *user_data)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    mosq_cb.on_message = cb_message;
    mosq_cb.on_message_user_data = user_data;
    return 0;
#else
    return -1;
#endif
}

/**
 * @brief   MQTT客户端 设置 订阅ACK的回调函数
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @param[in] cb_subscribe: MQTT 收到订阅ACK后的回调函数
 * @param[in] user_data: 用户数据
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_set_cb_subscribe(jl_mqtt_t mqtt_handler, int32_t (*cb_subscribe)(jl_mqtt_t, int32_t, void *), void *user_data)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    mosq_cb.on_subscribe = cb_subscribe;
    mosq_cb.on_subscribe_user_data = user_data;
    return 0;
#else
    return -1;
#endif
}

/**
 * @brief   MQTT客户端 设置 消息发布ACk的回调函数
 *
 * @param[in] mqtt_handler: MQTT 客户端句柄
 * @param[in] cb_publish: MQTT 收到消息发送ACK后的回调函数
 * @param[in] user_data: 用户数据
 * @return 
 	返回值 == 0,成功
    返回值 >< 0,失败 
 * @see None.
 * @note None.
 */
int32_t jl_platform_mqtt_set_cb_publish(jl_mqtt_t mqtt_handler, int32_t (*cb_publish)(jl_mqtt_t, int32_t, void *), void *user_data)
{
#ifdef __MQTT_MOSQUITTO_PAL__
    mosq_cb.on_publish = cb_publish;
    mosq_cb.on_publish_user_data = user_data;
    return 0;
#else
    return -1;
#endif
}

// end of file
